# -*- coding: utf-8 -*-
"""
Container for all built-in visualization modules.
"""

from mesa.visualization.modules.CanvasGridVisualization import CanvasGrid  # flake8: noqa
from mesa.visualization.modules.ChartVisualization import ChartModule  # flake8: noqa
from mesa.visualization.modules.TextVisualization import TextElement  # flake8: noqa
